from configparser import ConfigParser
import os


# todo czy ten Config nie powinien rozszerzac klasy ConfigParser?
class Config:
    def __init__(self):
        self._config = ConfigParser()
        config_file = os.path.dirname(os.path.realpath('__main__')) + '/config.ini'
        self.check_config_file_exists(config_file)
        self._config.read(config_file)
        required_options = self.required_options
        self.check_required_config_options(self._config, required_options)

    @staticmethod
    def check_config_file_exists(config_file):
        if not os.path.exists(config_file):
            raise FileNotFoundError('File {config_file} does not exist.'.format(config_file=config_file))

    @property
    def required_options(self):
        return {
            'facebook': [
                {'name': 'SENDER_ACTION',
                 'type': bool},
                {'name': 'FB_VERIFICATION_TOKEN',
                 'type': str},
                {'name': 'FB_TOKEN',
                 'type': str},
                {'name': 'FB_API',
                 'type': str},
                {'name': 'FB_GRAPH',
                 'type': str}
            ],
            'app': [
                {'name': 'PORT',
                 'type': int},
                {'name': 'LOG_FILE_NAME',
                 'type': str},
                {'name': 'LOG_TO_TERMINAL',
                 'type': bool}
            ]
        }

    def check_required_config_options(self, config, required_options):
        if not isinstance(config, ConfigParser):
            raise TypeError(
                'Invalid config type. Expected {expected}, {given} given.'.format(expected=ConfigParser.__name__,
                                                                                  given=type(config)))
        for section in required_options:
            if not config.has_section(section):
                self.raise_missing_section_error(section)
            for option in required_options.get(section):
                option_name = option.get('name')
                if not config.has_option(section, option_name):
                    self.raise_missing_option_error(option, section)
                self.check_option_type(section, option)

    def check_option_type(self, section, option):
        option_name = option.get('name')
        option_type = option.get('type')
        if not option_name or not option_type:
            raise ValueError(
                'Missing section or option in function: {function}'.format(function=self.check_option_type.__name__))
        if option_type is str:
            if not self.get_parameter(section, option_name):
                raise self.raise_missing_option_error(option_name, section)
        elif option_type is int:
            try:
                parameter = self.get_int_parameter(section, option_name)
            except ValueError:
                raise ValueError(
                    '[{section}]{option} is not integer.'.format(option=option_name, section=section))
            if not parameter:
                raise self.raise_missing_option_error(option_name, section)
        elif option_type is bool:
            try:
                self.get_bool_parameter(section, option_name)
            except ValueError:
                raise ValueError(
                    '[{section}]{option} is not boolean.'.format(option=option_name, section=section))
        else:
            raise TypeError(
                'Unsupported config value type: [{section}]{option}.'.format(section=section, option=option_name))

    @staticmethod
    def raise_missing_section_error(section):
        raise ValueError('Missing config section:{section}.'.format(section=section))

    @staticmethod
    def raise_missing_option_error(option, section):
        raise ValueError(
            'Missing config option: [{section}]{option}.'.format(option=option, section=section))

    @property
    def config(self):
        return self._config

    def get_parameter(self, section, option):
        parameter = self._config[section][option]
        if not parameter:
            self.raise_missing_option_error(option, section)
        return parameter

    def get_int_parameter(self, section, option):
        parameter = self._config.getint(section, option)
        if not parameter:
            self.raise_missing_option_error(option, section)
        return parameter

    def get_bool_parameter(self, section, option):
        return self._config.getboolean(section, option)



import nltk


class TextPaginator(object):
    def __init__(self, text: str,
                 message_threshold=140,
                 page_length_limit=320):
        self._pages = list()
        self._tokenizer = nltk.data.load('nltk:tokenizers/punkt/polish.pickle')
        self._sentences = self._tokenizer.tokenize(text)
        self._page_length_limit = page_length_limit
        self._message_threshold = message_threshold

        self._paginate()

    def get_page(self, page: int):
        return self._pages[page - 1]

    @property
    def pages_count(self):
        return len(self._pages)

    @property
    def sentences_count(self):
        return len(self._sentences)

    @property
    def page_length_limit(self):
        return self._page_length_limit

    @property
    def message_threshold(self):
        return self._message_threshold

    @property
    def sentences(self):
        return self._sentences

    def _paginate(self):
        page_content = ''
        for index, sentence in enumerate(self.sentences):
            if len(page_content + sentence) <= self.page_length_limit:
                if len(page_content):
                    sentence = ' ' + sentence
                page_content += sentence
            elif len(page_content) > self.message_threshold and len(sentence) <= self.page_length_limit:
                self._pages.append(page_content)
                page_content = sentence
            else:
                sentence_split_index = 0
                while True:
                    end_index_in_sentence = self.page_length_limit - len(page_content) + sentence_split_index
                    page_content += sentence[sentence_split_index: end_index_in_sentence]
                    sentence_split_index = end_index_in_sentence
                    if len(page_content) == self.page_length_limit:
                        self._pages.append(page_content)
                        page_content = ''
                    else:
                        break
            if index == self.sentences_count - 1 and page_content:
                self._pages.append(page_content)



from jellyfish import jaro_distance
import logging


class SimilaritySort(object):
    @staticmethod
    def _calculate_similarity(data: list, value, sorting_property=None):
        similarity_list = list()
        for entry in data:
            if isinstance(entry, str):
                if sorting_property is not None:
                    logging.warning('"sorting_property" is not necessary in string sorting.')
                attr = entry
            else:
                if sorting_property is None:
                    raise ValueError('Sorting property must be defined.')
                sorting_property = str(sorting_property)
                if isinstance(entry, dict):
                    if sorting_property not in entry:
                        raise ValueError(
                            'Key {property} is missing in given dict.'.format(property=sorting_property))
                    attr = entry.get(sorting_property)
                elif isinstance(entry, list):
                    raise NotImplementedError()
                else:
                    attr = getattr(entry, sorting_property)
                if attr is None:
                    attr = ''
            similarity_list.append(jaro_distance(str(attr), str(value)))

        return similarity_list

    @staticmethod
    def sorted_by(data: list, value, sorting_property=None, similarity_threshold=0, reverse=True):
        if not isinstance(reverse, bool):
            raise ValueError('Reverse value must be type of boolean.')
        similarity_list = SimilaritySort._calculate_similarity(data=data,
                                                               sorting_property=sorting_property,
                                                               value=value)
        zipped = zip(data, similarity_list)
        sorted_data = sorted(zipped, key=lambda k: k[1], reverse=reverse)

        data = list()
        for entry in sorted_data:
            if entry[1] >= similarity_threshold:
                data.append(entry[0])

        return data



from html.parser import HTMLParser


class HTMLStripper(HTMLParser):
    def __init__(self):
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.fed = []

    def handle_data(self, d):
        self.fed.append(d)

    def get_data(self):
        return ''.join(self.fed)

    @staticmethod
    def _strip_tags(html):
        s = HTMLStripper()
        s.feed(html)
        return s.get_data()

    @staticmethod
    def strip(data):
        if isinstance(data, str):
            return HTMLStripper._strip_tags(data)
        elif isinstance(data, dict):
            for key, value in data.items():
                value = HTMLStripper.strip(value)
                data.update({
                    key: value
                })
        elif isinstance(data, list):
            for item in data:
                HTMLStripper.strip(item)

        return data



class FbResponse:
    def __init__(self, template=None):
        self._send = False
        self._templates = list()
        if template:
            self.add_template(template)

    def add_template(self, template):
        self._templates.append(template)

    @property
    def templates(self):
        return self._templates

    @property
    def send(self):
        return self._send

    @send.setter
    def send(self, condition: bool):
        self._send = condition



from datetime import datetime, timedelta
import logging

from tornado import escape, gen, httpclient
from urllib import parse
import validators


class ApiClient:
    def __init__(self):
        self._http = httpclient.AsyncHTTPClient()

    @gen.coroutine
    def async_request(self, url, parameters=None, http_method='GET', body=None, headers=None):
        url = self.encode_url(url, parameters)
        default_header = {'Content-Type': 'application/json'}
        headers = self.append_headers(header_to_append=default_header, base_headers=headers)

        if body:
            body = self._validate_body(body)
        if not self._is_method_valid(http_method):
            raise ValueError('Unsupported method: {method}'.format(method=http_method))
        if not self._is_headers_type_valid(headers):
            raise TypeError('Invalid headers type')

        try:
            logging.debug('Requesting: {url}'.format(url=url))
            response = yield self._http.fetch(httpclient.HTTPRequest(url, http_method, headers, body=body))
            logging.debug('Success: {url}'.format(url=url))
            if response.body:
                return escape.json_decode(response.body)
        except httpclient.HTTPError as e:
            e.message = 'Unable to complete {url}: {message}'.format(url=url, message=e.message)
            raise e

    def _is_method_valid(self, http_method):
        supported_methods = ['GET', 'POST', 'DELETE', 'PUT', 'PATCH']
        return http_method in supported_methods

    @staticmethod
    def encode_url(url, parameters=None):
        if parameters:
            if not isinstance(parameters, dict):
                raise TypeError('Parameters must be instance of dict.')
            url += parse.urlencode(parameters)
        if not validators.url(url):
            raise ValueError('Invalid url format: {url}'.format(url=url))
        return url

    def _is_json_valid(self, json):
        is_valid = True
        try:
            escape.json_decode(json)
        except ValueError:
            is_valid = False
        return is_valid

    def _validate_body(self, body):
        if isinstance(body, dict):
            body = escape.json_encode(body)
        elif not self._is_json_valid(body):
            raise ValueError('Invalid json format')
        return body

    def _is_headers_type_valid(self, headers):
        return bool(isinstance(headers, dict))

    def append_headers(self, header_to_append: dict, base_headers=None):
        if base_headers is None:
            base_headers = dict()
        else:
            if not isinstance(base_headers, dict):
                raise TypeError(
                    'Invalid type in function append_headers, expected {expected}, {given} given.'.format(
                        expected=dict.__name__,
                        given=type(base_headers)))
        z = base_headers.copy()
        z.update(header_to_append)
        return z


class OAuthApiClient(ApiClient):
    def __init__(self,
                 oauth_authorization_url,
                 oauth_client_id,
                 oauth_client_secret,
                 oauth_grant_type,
                 username=None,
                 password=None,
                 renew_token_before_seconds=None,
                 non_expiring_token=None):
        super().__init__()
        self._access_token = None
        self._token_expires_at = None
        self._oauth_authorization_url = oauth_authorization_url
        self._oauth_client_id = oauth_client_id
        self._oauth_client_secret = oauth_client_secret
        self._oauth_grant_type = oauth_grant_type
        self._username = username
        self._password = password
        self._renew_token_before_seconds = renew_token_before_seconds
        self._non_expiring_token = non_expiring_token

    @gen.coroutine
    def oauth_async_request(self, url, parameters=None, http_method='GET', body=None, headers=None):
        access_token = yield self._get_access_token()
        auth_header = {'Authorization': 'Bearer {access_token}'.format(access_token=access_token)}
        headers = self.append_headers(header_to_append=auth_header, base_headers=headers)
        response = yield self.async_request(url=url, parameters=parameters, http_method=http_method, body=body, headers=headers)
        return response

    @gen.coroutine
    def _get_access_token(self):
        if self._renew_token_before_seconds is None:
            self._renew_token_before_seconds = 300

        if self._non_expiring_token is not None:
            return self._non_expiring_token

        if self._access_token is None:
            yield self._oauth_authorization_request(self._oauth_grant_type)
        elif (self._token_expires_at - datetime.now()).total_seconds() < int(self._renew_token_before_seconds):
            if 'refresh_token' in self._get_supported_grant_types():
                yield self._oauth_authorization_request('refresh_token')
            else:
                yield self._oauth_authorization_request(self._oauth_grant_type)
        return str(self._access_token)

    def _set_access_token(self, access_token):
        self._access_token = access_token

    def _set_access_token_expires_at(self, expires_in: int):
        self._token_expires_at = datetime.now() + timedelta(seconds=expires_in)

    @gen.coroutine
    def _oauth_authorization_request(self, grant_type, method='POST'):
        url = self._oauth_authorization_url
        payload = self._get_payload(grant_type)
        response = yield self.async_request(url, method, payload)
        self._set_access_token(response.get('access_token'))
        self._set_access_token_expires_at(int(response.get('expires_in')))

    def _get_payload(self, grant_type):
        if not self._is_grant_type_valid(grant_type):
            raise ValueError('Unsupported grant type: {grant_type}'.format(grant_type=grant_type))
        payload = {'client_id': self._oauth_client_id, 'client_secret': self._oauth_client_secret}
        if grant_type == 'password':
            payload.update({'username': self._username, 'password': self._password})
        elif grant_type == 'refresh_token':
            current_access_token = self._access_token
            if not current_access_token:
                raise ValueError('Cannot refresh empty token: {current_access_token}.'.format(
                    current_access_token=current_access_token))
            payload.update({'refresh_token': current_access_token})
        payload.update({'grant_type': grant_type})
        return payload

    @staticmethod
    def _get_supported_grant_types():
        return ['password', 'client_credentials']

    def _is_grant_type_valid(self, grant_type):
        is_valid = False
        supported_grant_types = self._get_supported_grant_types()
        if not isinstance(supported_grant_types, list):
            raise TypeError(
                'Function {function} must return list.'.format(function=self._get_supported_grant_types.__name__))
        if grant_type in supported_grant_types:
            is_valid = True
        return is_valid



import logging

from tornado import gen

from messenger_sdk.events import *


class Processor:
    __metaclass__ = ABCMeta

    @gen.coroutine
    def handle(self, event: Event):
        if not isinstance(event, Event):
            raise TypeError(
                'Invalid event class. Expected {}, {} given'.format(Event.__name__, type(event).__name__))
        response = yield self.process_event(event)
        return response

    @gen.coroutine
    def process_event(self, event: Event):
        response = yield self.handle_event(event)
        return response

    @gen.coroutine
    def handle_event(self, event: Event):
        intent = event.intent()
        if not intent:
            raise ValueError('Missing event intent')
        function_to_call = self._get_function_to_call(intent)
        if function_to_call:
            logging.debug('Function: %s', function_to_call.__name__)
            response = yield function_to_call(event)
            return response

    @abstractmethod
    def _get_function_to_call(self, intent):
        raise NotImplementedError()



# import requests
# from messenger_sdk.templates import BaseTemplate
#
#
# class Api:
#     def __init__(self, access_token, api_version=2.6):
#         self.access_token = access_token
#         self.api_version = api_version
#         self.endpoint = "https://graph.facebook.com/v{0}/me/messages?access_token={1}".format(
#             self.api_version,
#             self.access_token
#         )
#
#     def send_message(self, message: BaseTemplate):
#         return self._send_message_request(message.get_template())
#
#     def _send_message_request(self, payload):
#         response = requests.post(self.endpoint, json=payload).json()
#
#         return response
